from __future__ import annotations

import itertools
from collections.abc import Mapping, Sequence
from typing import Any, TypeGuard

import stim

from tqec.circuit.qubit_map import QubitMap
from tqec.circuit.schedule.circuit import ScheduledCircuit
from tqec.compile.blocks.layers.atomic.base import BaseLayer
from tqec.compile.blocks.layers.atomic.layout import LayoutLayer
from tqec.compile.blocks.layers.atomic.plaquettes import PlaquetteLayer
from tqec.compile.blocks.layers.atomic.raw import RawCircuitLayer
from tqec.compile.blocks.layers.composed.base import BaseComposedLayer
from tqec.compile.blocks.layers.composed.repeated import RepeatedLayer
from tqec.compile.blocks.layers.composed.sequenced import SequencedLayers
from tqec.compile.tree.annotations import LayerNodeAnnotations, Polygon
from tqec.utils.coordinates import StimCoordinates
from tqec.utils.exceptions import TQECError
from tqec.utils.scale import LinearFunction


def contains_only_layout_or_composed_layers(
    layers: Sequence[BaseLayer | BaseComposedLayer],
) -> TypeGuard[Sequence[LayoutLayer | BaseComposedLayer]]:
    """Ensure correct typing after using that function in a condition."""
    return all(isinstance(layer, (LayoutLayer, BaseComposedLayer)) for layer in layers)


class NodeWalker:
    def visit_node(self, node: LayerNode) -> None:
        """Interface called when ``node`` is visited, before recursing in children."""
        pass  # pragma: no cover

    def enter_node(self, node: LayerNode) -> None:
        """Interface called when entering ``node``."""
        pass

    def exit_node(self, node: LayerNode) -> None:
        """Interface called when exiting ``node``."""
        pass


class LayerNode:
    def __init__(
        self,
        layer: LayoutLayer | BaseComposedLayer,
        annotations: Mapping[int, LayerNodeAnnotations] | None = None,
    ) -> None:
        """Represent a node in a :class:`~tqec.compile.tree.tree.LayerTree`.

        Args:
            layer: layer being represented by the node.
            annotations: already computed annotations. Default to ``None`` meaning
                no annotations are provided. Should be a mapping from values of
                ``k`` to the annotations already computed for that value of ``k``.

        """
        self._layer = layer
        self._children = LayerNode._get_children(layer)
        self._annotations = dict(annotations) if annotations is not None else {}

    @staticmethod
    def _get_children(layer: LayoutLayer | BaseComposedLayer) -> list[LayerNode]:
        if isinstance(layer, LayoutLayer):
            return []
        if isinstance(layer, SequencedLayers):
            if not contains_only_layout_or_composed_layers(layer.layer_sequence):
                raise TQECError(
                    "Found a leaf node that is not an instance of "
                    f"{LayoutLayer.__name__}. This should not happen and is a "
                    "logical error."
                )
            return [LayerNode(lay) for lay in layer.layer_sequence]  # type: ignore
        if isinstance(layer, RepeatedLayer):
            if not isinstance(layer.internal_layer, LayoutLayer | BaseComposedLayer):
                raise TQECError(
                    "The layer that is being repeated is not an instance of "
                    f"{LayoutLayer.__name__} or {BaseComposedLayer.__name__}."
                )
            return [LayerNode(layer.internal_layer)]  # type: ignore
        if isinstance(layer, (PlaquetteLayer, RawCircuitLayer)):
            raise TQECError(
                f"Unsupported layer type found: {type(layer).__name__}. Expected "
                f"ALL leaf nodes to be of type {LayoutLayer.__name__}."
            )
        raise NotImplementedError(f"Unknown layer type found: {type(layer).__name__}.")

    @property
    def is_leaf(self) -> bool:
        """Return ``True`` if ``self`` does not have any children and so is a leaf node."""
        return isinstance(self._layer, LayoutLayer)

    @property
    def is_repeated(self) -> bool:
        """Return ``True`` if ``self`` stores a RepeatedLayer."""
        return isinstance(self._layer, RepeatedLayer)

    @property
    def repetitions(self) -> LinearFunction | None:
        """Return the number of repetitions of the node if ``self.is_repeated`` else ``None``."""
        return self._layer.repetitions if isinstance(self._layer, RepeatedLayer) else None

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of ``self``."""
        return {  # pragma: no cover
            "layer": type(self._layer).__name__,
            "children": [child.to_dict() for child in self._children],
            "annotations": {k: annotation.to_dict() for k, annotation in self._annotations.items()},
        }

    def walk(self, walker: NodeWalker) -> None:
        """Walk the tree using Depth-First Search (DFS), calling walker methods on each node.

        Args:
            walker: structure that will be called on each explored node.

        """
        walker.enter_node(self)
        walker.visit_node(self)
        for child in self._children:
            child.walk(walker)
        walker.exit_node(self)

    @property
    def children(self) -> list[LayerNode]:
        """Return the children nodes of ``self``."""
        return self._children

    def get_annotations(self, k: int) -> LayerNodeAnnotations:
        """Return the annotations associated with the provided scaling parameter ``k``."""
        return self._annotations.setdefault(k, LayerNodeAnnotations())

    def set_circuit_annotation(self, k: int, circuit: ScheduledCircuit) -> None:
        """Set the circuit annotation associated with the scaling parameter ``k`` to ``circuit``."""
        self.get_annotations(k).circuit = circuit

    def generate_circuits_with_potential_polygons(
        self,
        k: int,
        global_qubit_map: QubitMap,
        add_polygons: bool = False,
    ) -> list[stim.Circuit | list[Polygon]]:
        """Generate the circuits and polygons for each nodes in the subtree rooted at ``self``.

        Args:
            k: scaling parameter.
            global_qubit_map: qubit map that should be used to generate the
                quantum circuit. Qubits from the returned quantum circuit will
                adhere to the provided qubit map.
            add_polygons: if ``True``, polygon objects for visualization in Crumble
                will be added to the returned list.

        Returns:
            a list of ``stim.Circuit`` and/or ``list[Polygon]`` objects.
            Each ``stim.Circuit`` represents a quantum circuit of a leaf node in
            the tree. Each polygon list represents the stabilizer configuration
            for the corresponding leaf node and will be placed right before the
            corresponding circuit in the returned list. If two consecutive leaf
            nodes have the same stabilizer configuration, only the first polygons
            will be kept.

        """
        if isinstance(self._layer, LayoutLayer):
            annotations = self.get_annotations(k)
            base_circuit = annotations.circuit
            if base_circuit is None:
                raise TQECError(
                    "Cannot generate the final quantum circuit before annotating "
                    "nodes with their individual circuits. Did you call "
                    "LayerTree.annotate_circuits before?"
                )
            local_qubit_map = base_circuit.qubit_map
            qubit_indices_mapping = {
                local_qubit_map[q]: global_qubit_map[q] for q in local_qubit_map.qubits
            }
            mapped_circuit = base_circuit.map_qubit_indices(qubit_indices_mapping)
            for annotation in annotations.detectors + annotations.observables:
                mapped_circuit.append_annotation(annotation.to_instruction())
            mapped_circuit.append_annotation(
                stim.CircuitInstruction(
                    "SHIFT_COORDS", [], StimCoordinates(0, 0, 1).to_stim_coordinates()
                )
            )
            ret: list[stim.Circuit | list[Polygon]] = [
                mapped_circuit.get_circuit(include_qubit_coords=False)
            ]
            if add_polygons:
                ret.insert(0, annotations.polygons)

            return ret

        if isinstance(self._layer, SequencedLayers):
            ret = []
            for child, next_child in itertools.pairwise(self._children):
                ret += child.generate_circuits_with_potential_polygons(
                    k, global_qubit_map, add_polygons
                )
                if not next_child.is_repeated:
                    assert isinstance(ret[-1], stim.Circuit)
                    ret[-1].append("TICK", [], [])
            ret += self._children[-1].generate_circuits_with_potential_polygons(
                k, global_qubit_map, add_polygons
            )
            return ret

        if isinstance(self._layer, RepeatedLayer):
            body = self._children[0].generate_circuits_with_potential_polygons(
                k, global_qubit_map, add_polygons=add_polygons
            )
            body_circuit = sum(
                (i for i in body if isinstance(i, stim.Circuit)),
                start=stim.Circuit(),
            )
            body_circuit.insert(0, stim.CircuitInstruction("TICK"))
            ret = []
            if add_polygons:
                # only keep the first set of polygons
                ret.append(body[0])
            ret.append(body_circuit * self._layer.repetitions.integer_eval(k))
            return ret
        raise TQECError(f"Unknown layer type found: {type(self._layer).__name__}.")

    def generate_circuit(self, k: int, global_qubit_map: QubitMap) -> stim.Circuit:
        """Generate the quantum circuit representing the node.

        Args:
            k: scaling parameter.
            global_qubit_map: qubit map that should be used to generate the
                quantum circuit. Qubits from the returned quantum circuit will
                adhere to the provided qubit map.

        Returns:
            a ``stim.Circuit`` instance representing ``self`` with the provided
            ``global_qubit_map``.

        """
        circuits = self.generate_circuits_with_potential_polygons(
            k, global_qubit_map, add_polygons=False
        )
        ret = stim.Circuit()
        for circuit in circuits:
            assert isinstance(circuit, stim.Circuit)
            ret += circuit
        return ret



from abc import ABC, abstractmethod

from tqec.plaquette.plaquette import Plaquette
from tqec.plaquette.rpng import RPNGDescription


class RPNGTranslator(ABC):
    """Base interface for classes capable of generating a plaquette from a RPNG description."""

    @abstractmethod
    def translate(
        self,
        rpng_description: RPNGDescription,
    ) -> Plaquette:
        """Generate the plaquette corresponding to the provided ``RPNG`` description.

        Args:
            rpng_description: description of the plaquette to generate.

        Returns:
            a valid implementation of the provided
            :class:`~tqec.plaquette.rpng.RPNGDescription` instance.

        """
        pass



from __future__ import annotations

from collections.abc import Callable, Iterable

from typing_extensions import override

from tqec.circuit.schedule.circuit import ScheduledCircuit
from tqec.compile.blocks.enums import SpatialBlockBorder
from tqec.compile.blocks.layers.atomic.base import BaseLayer
from tqec.utils.scale import LinearFunction, PhysicalQubitScalable2D


class RawCircuitLayer(BaseLayer):
    def __init__(
        self,
        circuit_factory: Callable[[int], ScheduledCircuit],
        scalable_raw_shape: PhysicalQubitScalable2D,
        scalable_num_moments: LinearFunction,
        trimmed_spatial_borders: frozenset[SpatialBlockBorder] = frozenset(),
    ):
        """Represent a layer with a spatial footprint that is defined by a raw circuit.

        Args:
            circuit_factory: a function callable returning a quantum circuit for
                any input ``k >= 1``.
            scalable_raw_shape: scalable shape of the quantum circuit returned
                by the provided ``circuit_factory``.
            scalable_num_moments: a linear function associating to any input
                ``k >= 1`` the number of moments returned by the provided
                ``circuit_factory`` when given ``k`` as input. This is expected
                to be constant in most scenario, but left as a linear function
                to cover potential edge-cases.
            trimmed_spatial_borders: all the spatial borders that have been
                removed from the layer.

        """
        super().__init__(trimmed_spatial_borders)
        self._circuit_factory = circuit_factory
        self._scalable_raw_shape = scalable_raw_shape
        self._scalable_num_moments = scalable_num_moments

    @property
    def scalable_raw_shape(self) -> PhysicalQubitScalable2D:
        """Get the scalable shape of the quantum circuit returned by ``self.circuit_factory``."""
        return self._scalable_raw_shape

    @property
    def circuit_factory(self) -> Callable[[int], ScheduledCircuit]:
        """Get the callable used to generate a scalable quantum circuit from the scaling factor."""
        return self._circuit_factory  # pragma: no cover

    @property
    @override
    def scalable_shape(self) -> PhysicalQubitScalable2D:
        return self.scalable_raw_shape

    @override
    def with_spatial_borders_trimmed(
        self, borders: Iterable[SpatialBlockBorder]
    ) -> RawCircuitLayer:
        raise NotImplementedError(
            f"Cannot trim spatial borders of a {RawCircuitLayer.__name__} instance."
        )

    @override
    def __eq__(self, value: object) -> bool:
        raise NotImplementedError()

    def __hash__(self) -> int:
        raise NotImplementedError(f"Cannot hash efficiently a {type(self).__name__}.")

    @property
    @override
    def scalable_num_moments(self) -> LinearFunction:
        return self._scalable_num_moments  # pragma: no cover



"""Interoperability with COLLADA DAE files."""

from tqec.interop.collada.html_viewer import (
    display_collada_model as display_collada_model,
)
from tqec.interop.collada.read_write import (
    read_block_graph_from_dae_file as read_block_graph_from_dae_file,
)
from tqec.interop.collada.read_write import (
    write_block_graph_to_dae_file as write_block_graph_to_dae_file,
)



from collections.abc import Callable, Sequence
from dataclasses import dataclass, field

import stim
from typing_extensions import override

from tqec.circuit.moment import Moment
from tqec.circuit.schedule.circuit import ScheduledCircuit
from tqec.plaquette.compilation.passes.base import CompilationPass
from tqec.plaquette.compilation.passes.transformer.schedule import ScheduleFunction
from tqec.plaquette.compilation.passes.transformer.simplifiers import (
    InstructionSimplifier,
    NoInstructionSimplification,
)


@dataclass
class InstructionCreator:
    """Create an instruction from targets and arguments."""

    name: str
    targets: Callable[[list[stim.GateTarget]], list[stim.GateTarget]] = lambda trgts: trgts
    arguments: Callable[[list[float]], list[float]] = lambda args: args

    def __call__(
        self, targets: list[stim.GateTarget], arguments: list[float]
    ) -> stim.CircuitInstruction:
        """Create a ``stim.CircuitInstruction`` from the provided arguments."""
        return stim.CircuitInstruction(self.name, self.targets(targets), self.arguments(arguments))


@dataclass
class ScheduledCircuitTransformation:
    """Describes an instruction transformation.

    This class describes how a given instruction should be transformed into
    potentially several instructions.

    Attributes:
        source_name: name of the instruction that is transformed by ``self``.
        transformation: a mapping from a schedule description to a list of
            instruction creators. The instructions created by the instruction
            creators will be inserted at the schedule computed by the schedule
            description.
        instruction_simplifier: a simplifier applied before trying to create a
            :class:`~tqec.circuit.moment.Moment` instance with the instructions
            resulting from the application of ``self``.

    """

    source_name: str
    transformation: dict[ScheduleFunction, list[InstructionCreator]]
    instruction_simplifier: InstructionSimplifier = field(
        default_factory=NoInstructionSimplification
    )

    def apply(self, circuit: ScheduledCircuit) -> ScheduledCircuit:
        """Apply the transformation to ``circuit`` and return the result."""
        # moment_instructions: schedule_index -> instruction list.
        moment_instructions: dict[int, list[stim.CircuitInstruction]] = {}
        for schedule, moment in circuit.scheduled_moments:
            for instruction in moment.instructions:
                # if the transformation represented by self does not apply to
                # the current instruction, just add it unmodified.
                if instruction.name not in stim.gate_data(self.source_name).aliases:
                    moment_instructions.setdefault(schedule, []).append(instruction)
                    continue
                # else, for each instruction creator in self.transformation, add
                # the created instruction to the target moment.
                targets = instruction.targets_copy()
                args = instruction.gate_args_copy()
                for schedule_function, instr_creators in self.transformation.items():
                    sched: int = schedule_function(schedule)
                    moment_instructions.setdefault(sched, []).extend(
                        creator(targets, args) for creator in instr_creators
                    )
        # Make sure that the schedules are given to ScheduledCircuit as a
        # sorted list.
        schedules = sorted(moment_instructions.keys())
        all_moments = [
            Moment.from_instructions(
                # Try to simplify operations before creating the moment.
                self.instruction_simplifier.simplify(moment_instructions[s])
            )
            for s in schedules
        ]
        return ScheduledCircuit(all_moments, schedules, circuit.qubit_map)


class ScheduledCircuitTransformer:
    def __init__(self, transformations: Sequence[ScheduledCircuitTransformation]) -> None:
        """Describe a list of :class:`ScheduledCircuitTransformation` instances.

        Note:
            This class has been introduced for convenience and for future
            optimisation. Right now, a new scheduled circuit is created for each
            :class:`ScheduledCircuitTransformation` instance in ``self``. This is
            suboptimal as we might be able to apply all the transformations by
            iterating the original quantum circuit once.

            Due to the very limited size of the circuits given to the compilation
            pipeline, this performance issue does not seem to have a measurable
            impact at the moment.

        """
        self._transformations = transformations

    def apply(self, circuit: ScheduledCircuit) -> ScheduledCircuit:
        """Apply the transformations stored in ``self`` to ``circuit`` and return the result."""
        for transformation in self._transformations:
            circuit = transformation.apply(circuit)
        return circuit


class ScheduledCircuitTransformationPass(CompilationPass):
    def __init__(
        self,
        transformations: Sequence[ScheduledCircuitTransformation],
    ) -> None:
        """Apply the provided transformations as a compilation pass.

        Args:
            transformations: a sequence of transformation that are applied one after the other.

        """
        super().__init__()
        self._transformations = ScheduledCircuitTransformer(transformations)

    @override
    def run(self, circuit: ScheduledCircuit, check_all_flows: bool = False) -> ScheduledCircuit:
        modified_circuit = self._transformations.apply(circuit)
        if check_all_flows:
            self.check_flows(circuit, modified_circuit)  # pragma: no cover
        return modified_circuit



from __future__ import annotations

from collections.abc import Mapping

from typing_extensions import override

from tqec.compile.blocks.enums import TemporalBlockBorder
from tqec.compile.blocks.layers.spatial import WithSpatialFootprint
from tqec.compile.blocks.layers.temporal import WithTemporalFootprint
from tqec.utils.exceptions import TQECError
from tqec.utils.scale import LinearFunction


class BaseLayer(WithSpatialFootprint, WithTemporalFootprint):
    """Base class representing a "layer".

    A "layer" is a scalable and repeatable unit of a quantum error correction
    circuit. In most cases, it corresponds to a single round of the syndrome
    extraction circuit. A base layer occupies a specific spacetime footprint,
    where the spatial footprint scales with the layer circuit, while the temporal
    footprint remains fixed at ``1``, representing a single layer. To stack layers
    over time or repeat them in a scalable manner, use
    :class:`~tqec.compile.blocks.layers.composed.base.BaseComposedLayer` instead.

    """

    @property
    @override
    def scalable_timesteps(self) -> LinearFunction:
        # By definition of a "layer":
        return LinearFunction(0, 1)

    @override
    def with_temporal_borders_replaced(
        self, border_replacements: Mapping[TemporalBlockBorder, BaseLayer | None]
    ) -> BaseLayer | None:
        if not border_replacements:
            return self
        if len(border_replacements) > 1 and any(
            replacement is not None for replacement in border_replacements.values()
        ):
            raise TQECError(
                "Unclear semantic: trying to replace the two temporal borders of "
                "an atomic layer that, by definition, only contain one layer."
            )
        return next(iter(border_replacements.values()))

    @override
    def get_temporal_layer_on_border(self, border: TemporalBlockBorder) -> BaseLayer:
        return self



from tqec.plaquette.compilation.passes.transformer import (
    InstructionCreator,
    ScheduledCircuitTransformation,
    ScheduledCircuitTransformationPass,
    ScheduleOffset,
)
from tqec.utils.enums import Basis


class ChangeMeasurementBasisPass(ScheduledCircuitTransformationPass):
    def __init__(self, basis: Basis):
        """Change ``MX`` and ``MZ`` instructions to the provided ``basis``."""
        ibasis = Basis.X if basis == Basis.Z else Basis.Z
        transformations = [
            ScheduledCircuitTransformation(
                f"M{ibasis.value.upper()}",
                {
                    ScheduleOffset(-1): [InstructionCreator("H")],
                    ScheduleOffset(0): [InstructionCreator(f"M{basis.value.upper()}")],
                },
            )
        ]

        super().__init__(transformations)



import warnings

import stim

from tqec.circuit.qubit_map import QubitMap
from tqec.utils.exceptions import TQECWarning


def subcircuit_only_on_indices(
    circuit: stim.Circuit, qubit_indices: frozenset[int]
) -> stim.Circuit:
    """Return a new circuit with only operations applied on qubit targets in ``qubit_indices``.

    Note:
        Non-qubit targets are not considered for filtering. That means that an
        operation only containing non-qubit targets (e.g., ``DETECTOR``) will
        not be filtered out. That also means that an operation applied on a
        qubit and a non-qubit target (e.g. ``CX rec[-1] 1``) will be in the
        returned circuit only if the qubit target(s) it is applied to (in the
        previous example, ``1``) is in ``qubit_indices``.

    Args:
        circuit: original circuit that should be filtered.
        qubit_indices: indices of qubits to keep in the resulting circuit.

    Returns:
        a new quantum circuit, only containing qubits with the indices provided.

    """
    ret = stim.Circuit()
    for instruction in circuit:
        if isinstance(instruction, stim.CircuitRepeatBlock):
            filtered_body = subcircuit_only_on_indices(instruction.body_copy(), qubit_indices)
            if len(filtered_body) != 0:
                ret.append(stim.CircuitRepeatBlock(instruction.repeat_count, filtered_body))
        elif instruction.name == "TICK":
            # Unconditionally append TICK annotations.
            ret.append(instruction)
        else:
            targets: list[stim.GateTarget] = []
            for target_group in instruction.target_groups():
                if all(
                    t.qubit_value is None or t.qubit_value in qubit_indices for t in target_group
                ):
                    targets.extend(target_group)
                if any(t.is_measurement_record_target for t in target_group):
                    warnings.warn(
                        "Found a measurement record target when filtering a "
                        "circuit. Filtering *might* remove qubit targets from "
                        "measurements instructions, and so *might* change the "
                        "measurement record. Be aware that the resulting "
                        "circuit might be invalid due to that behaviour.",
                        TQECWarning,
                    )
            if targets:
                ret.append(
                    stim.CircuitInstruction(instruction.name, targets, instruction.gate_args_copy())
                )
    return ret


def subcircuit(
    circuit: stim.Circuit, minx: float, maxx: float, miny: float, maxy: float
) -> stim.Circuit:
    """Filter a given circuit based on qubit coordinates.

    Only qubits within the bounding box provided by ``(minx, miny)`` and
    ``(maxx, maxy)`` will be kept in the resulting circuit. Only operations
    only involving kept qubits will be kept in the resulting circuit.

    Warning:
        Minimum bounds (``minx`` and ``miny``) are inclusive, but maximum ones
        (``maxx`` and ``maxy``) are exclusive.

    Args:
        circuit: original circuit that should be filtered.
        minx: X-coordinate of the top-left corner of the bounding box. Inclusive.
        maxx: X-coordinate of the bottom-right corner of the bounding box.
            Exclusive.
        miny: Y-coordinate of the top-left corner of the bounding box. Inclusive.
        maxy: Y-coordinate of the topbottom-right corner of the bounding box.
            Exclusive.

    Returns:
        a new quantum circuit, only containing qubits within the provided
        bounding box.

    """
    qubit_map = QubitMap.from_circuit(circuit)
    indices_to_keep = frozenset(
        i for i, q in qubit_map.i2q.items() if ((minx <= q.x < maxx) and (miny <= q.y < maxy))
    )
    return subcircuit_only_on_indices(circuit, indices_to_keep)



"""Provides utility code to help building generators.

The main utility provided by this module at the moment is
:class:`PlaquetteMapper` that is a "Funcorator" (a "functor", i.e., a callable
object that may have a state, and a "decorator") that can be used to define
functions that return :class:`~tqec.plaquettes.plaquettes.Plaquettes` instances
from one that returns `FrozenDefaultDict[int, RPNGDescription]`.

"""

from collections.abc import Callable
from functools import wraps
from types import FunctionType
from typing import Final, ParamSpec, cast

from tqec.plaquette.compilation.base import IdentityPlaquetteCompiler, PlaquetteCompiler
from tqec.plaquette.plaquette import Plaquette, Plaquettes
from tqec.plaquette.rpng.rpng import RPNGDescription
from tqec.plaquette.rpng.translators.base import RPNGTranslator
from tqec.plaquette.rpng.translators.default import DefaultRPNGTranslator
from tqec.utils.exceptions import TQECError
from tqec.utils.frozendefaultdict import FrozenDefaultDict

P = ParamSpec("P")


class PlaquetteMapper:
    def __init__(
        self,
        translator: RPNGTranslator = DefaultRPNGTranslator(),
        compiler: PlaquetteCompiler = IdentityPlaquetteCompiler,
    ) -> None:
        """Wrap a translator and a compiler to ease plaquette generation."""
        self._translator = translator
        self._compiler = compiler

    def get_plaquette(self, description: RPNGDescription) -> Plaquette:
        """Successively call the translator and the compiler to return a plaquette."""
        return self._compiler.compile(self._translator.translate(description))

    def __call__(
        self,
        f: Callable[P, FrozenDefaultDict[int, RPNGDescription]],
    ) -> Callable[P, Plaquettes]:
        """Wrap the provided callable ``f`` to automatically get :class:`.Plaquette` instances.

        This method wraps correctly the provided function such that :meth:`get_plaquette` is called
        on each of the returned :class:`.RPNGDescription` instance to get its corresponding
        :class:`.Plaquette`.

        """
        func = cast(FunctionType, f)

        # The wraps decorator make sure that the original function name, module,
        # docstring, ... is correctly transmitted to the wrapper.
        @wraps(f)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> Plaquettes:
            return Plaquettes(f(*args, **kwargs).map_values(self.get_plaquette))

        # Because the function name have to change, we need to explicitly change
        # it here.
        wrapped_func_name = func.__name__
        expected_end = "_rpng_descriptions"
        if not wrapped_func_name.endswith(expected_end):
            raise TQECError(
                f"Cannot wrap function {func.__module__}.{func.__name__}: its name "
                f"does not end with '{expected_end}'."
            )
        wrapped_name = wrapped_func_name[: -len(expected_end)] + "_plaquettes"
        wrapper.__name__ = wrapped_name
        # Return the final function.
        return wrapper


default_plaquette_mapper: Final = PlaquetteMapper()



from tqec.plaquette.plaquette import Plaquette
from tqec.plaquette.rpng.rpng import RPNGDescription
from tqec.plaquette.rpng.translators.base import RPNGTranslator
from tqec.plaquette.rpng.translators.default import DefaultRPNGTranslator
from tqec.utils.enums import Basis


def make_surface_code_plaquette(
    basis: Basis,
    reset: Basis | None = None,
    measurement: Basis | None = None,
    translator: RPNGTranslator = DefaultRPNGTranslator(),
) -> Plaquette:
    b = basis.value.lower()
    r = reset.value.lower() if reset is not None else "-"
    m = measurement.value.lower() if measurement is not None else "-"
    sched = [1, 2, 3, 5] if basis == Basis.X else [1, 4, 3, 5]
    return translator.translate(
        RPNGDescription.from_string(" ".join(f"{r}{b}{s}{m}" for s in sched))
    )



r"""Defines a high-level template composing several sub-templates: :class:`LayoutTemplate`.

This module defines a :class:`~tqec.templates.base.Template` child class that
is able to represent several :class:`~tqec.templates.base.RectangularTemplate`
instances with the same scalable shape. Each of the managed
:class:`~tqec.templates.base.RectangularTemplate` instance is linked to a unique
2-dimensional position on an infinite 2-dimensional grid.

Example:
-------
A grid of :math:`2 \times 2` logical qubits can be represented with

.. code-block:: python

    from tqec.utils.position import BlockPosition2D
    from tqec.templates.layout import LayoutTemplate
    from tqec.templates.qubit import QubitTemplate

    qubit = QubitTemplate()
    grid = LayoutTemplate(
        {
            BlockPosition2D(0, 0): qubit,
            BlockPosition2D(0, 1): qubit,
            BlockPosition2D(1, 0): qubit,
            BlockPosition2D(1, 1): qubit,
        }
    )

Being a :class:`~tqec.templates.base.Template` instance, the above constructed
``grid`` can be displayed with

.. code-block:: python

    from tqec.templates.display import display_template

    display_template(qubit, 2)
    print("=" * 50)
    display_template(grid, 2)

which outputs

.. code-block:: text

    1   5   6   5   6   2
    7   9  10   9  10  11
    8  10   9  10   9  12
    7   9  10   9  10  11
    8  10   9  10   9  12
    3  13  14  13  14   4
    ==================================================
     1   5   6   5   6   2  29  33  34  33  34  30
     7   9  10   9  10  11  35  37  38  37  38  39
     8  10   9  10   9  12  36  38  37  38  37  40
     7   9  10   9  10  11  35  37  38  37  38  39
     8  10   9  10   9  12  36  38  37  38  37  40
     3  13  14  13  14   4  31  41  42  41  42  32
    15  19  20  19  20  16  43  47  48  47  48  44
    21  23  24  23  24  25  49  51  52  51  52  53
    22  24  23  24  23  26  50  52  51  52  51  54
    21  23  24  23  24  25  49  51  52  51  52  53
    22  24  23  24  23  26  50  52  51  52  51  54
    17  27  28  27  28  18  45  55  56  55  56  46

"""

from collections.abc import Mapping, Sequence
from copy import deepcopy

import numpy
import numpy.typing as npt
from typing_extensions import override

from tqec.plaquette.plaquette import Plaquette, Plaquettes
from tqec.templates.base import RectangularTemplate, Template
from tqec.utils.exceptions import TQECError
from tqec.utils.frozendefaultdict import FrozenDefaultDict
from tqec.utils.position import (
    BlockPosition2D,
    PlaquettePosition2D,
    PlaquetteShape2D,
    Shift2D,
)
from tqec.utils.scale import PlaquetteScalable2D


class LayoutTemplate(Template):
    def __init__(
        self,
        element_layout: dict[BlockPosition2D, RectangularTemplate],
        default_increments: Shift2D | None = None,
    ) -> None:
        """Implement a template representing a layout of other templates.

        Each element template in the layout is placed at a specific position in
        the 2D grid.

        Note:
            The provided template positions have only one restriction: two
            templates should not be at the same position.
            In particular, this class does not require that the provided
            template are spatially connected or entirely cover a rectangular
            portion of the 2 spatial dimensions.

        Args:
            element_layout: a dictionary with the position of the element
                templates in the layout as keys and the templates as values.
            default_increments: default increments between two plaquettes. Defaults
                to ``Displacement(2, 2)`` when ``None``.

        """
        super().__init__(default_increments)
        if not element_layout:
            raise TQECError("Cannot create a layout with an empty template map.")

        scalable_shapes = {template.scalable_shape for template in element_layout.values()}
        if len(scalable_shapes) != 1:
            raise TQECError("All templates in the layout should have the same scalable shape.")
        self._element_scalable_shape = scalable_shapes.pop()

        all_positions = list(element_layout.keys())
        min_x = min(position.x for position in all_positions)
        max_x = max(position.x for position in all_positions)
        min_y = min(position.y for position in all_positions)
        max_y = max(position.y for position in all_positions)
        # Shift the bounding box to the origin
        self._block_origin = BlockPosition2D(min_x, min_y)
        self._nx = max_x - min_x + 1
        self._ny = max_y - min_y + 1

        self._layout = deepcopy(element_layout)

    def get_indices_map_for_instantiation(
        self,
        instantiate_indices: Sequence[int] | None = None,
    ) -> dict[BlockPosition2D, dict[int, int]]:
        """Get one index map for each of the templates in the layout.

        This method is used internally by :meth:`LayoutTemplate.instantiate` and
        can be used externally to get a map from the instantiation indices used by
        each of the individual :class:`~tqec.templates.base.Template` instances
        stored in ``self`` to the global instantiation index used.

        Args:
            instantiate_indices: global indices used to instantiate ``self``.
                Have the same meaning and defaults as the ``plaquette_indices``
                parameter in :meth:`LayoutTemplate.instantiate`.
                Defaults to None.

        Returns:
            a map from positions to instantiation index maps. Each instantiation
            index map is a bijection from the template instantiation indices (going
            from ``1`` to ``template.expected_plaquettes_number``) to the global
            indices that are used in :meth:`LayoutTemplate.instantiate` (going
            from ``N`` to ``N + template.expected_plaquettes_number - 1`` where
            ``N`` depends on the other :class:`~tqec.templates.base.Template`
            instances managed by ``self``).

        """
        if instantiate_indices is None:
            instantiate_indices = list(range(1, self.expected_plaquettes_number + 1))
        index_count = 0
        indices_map: dict[BlockPosition2D, dict[int, int]] = {}
        for position, template in self._layout.items():
            indices_map[position] = {
                i + 1: instantiate_indices[i + index_count]
                for i in range(template.expected_plaquettes_number)
            }
            index_count += template.expected_plaquettes_number
        return indices_map

    def get_global_plaquettes(
        self, individual_plaquettes: Mapping[BlockPosition2D, Plaquettes]
    ) -> Plaquettes:
        """Merge ``individual_plaquettes`` into a single :class:`.Plaquettes` instance.

        The returned :class:`.Plaquettes` instance can then be used to instantiate ``self``.

        Raises:
            TQECError: if the provided ``individual_plaquettes`` does not
                cover all the :class:`~tqec.utils.position.BlockPosition2D` where
                there is a template in ``self``.
            TQECError: if the provided ``individual_plaquettes`` have
                different values for their ``default_factory``.

        Args:
            individual_plaquettes: a mapping from positions to plaquette
                collections. The keys should cover all the positions where a
                template is present in ``self``.

        Returns:
            a unique :class:`~tqec.plaquette.plaquette.Plaquettes` instance that
            can be used with ``self`` to generate a quantum circuit with
            :meth:`~tqec.compile.generation.generate_circuit`.

        """
        missing_positions = frozenset(self._layout.keys()) - frozenset(individual_plaquettes.keys())
        if missing_positions:
            raise TQECError(
                "The following expected positions were not found in the "
                f"provided individual_plaquettes: {missing_positions}."
            )
        default_values: list[Plaquette | None] = []
        index_maps = self.get_indices_map_for_instantiation()
        global_plaquettes: dict[int, Plaquette] = {}
        for position, index_map in index_maps.items():
            global_plaquettes |= {
                index_map[local_index]: plaquette
                for local_index, plaquette in individual_plaquettes[position].collection.items()
            }
            default_values.append(individual_plaquettes[position].collection.default_value)
        unique_default_values = frozenset(default_values)
        if len(unique_default_values) != 1:
            raise TQECError(
                "Found several different default factories: "
                f"{unique_default_values}. Cannot pick one for the merged "
                f"{Plaquettes.__name__} instance."
            )
        return Plaquettes(
            FrozenDefaultDict(global_plaquettes, default_value=next(iter(unique_default_values)))
        )

    @property
    @override
    def scalable_shape(self) -> PlaquetteScalable2D:
        """Return a scalable version of the template shape."""
        return PlaquetteScalable2D(
            self._nx * self._element_scalable_shape.x,
            self._ny * self._element_scalable_shape.y,
        )

    def element_shape(self, k: int) -> PlaquetteShape2D:
        """Return the uniform shape of the element templates."""
        return self._element_scalable_shape.to_shape_2d(k)

    @property
    @override
    def expected_plaquettes_number(self) -> int:
        """Return the number of plaquettes expected from the ``instantiate`` method.

        Returns:
            the number of plaquettes expected from the `instantiate` method.

        """
        return sum(template.expected_plaquettes_number for template in self._layout.values())

    @override
    def instantiate(
        self, k: int, plaquette_indices: Sequence[int] | None = None
    ) -> npt.NDArray[numpy.int_]:
        """Generate the numpy array representing the template.

        Args:
            k: scaling parameter used to instantiate the template.
            plaquette_indices: the plaquette indices that will be forwarded to
                the underlying Template instance's instantiate method. Defaults
                to `range(1, self.expected_plaquettes_number + 1)` if `None`.

        Returns:
            a numpy array with the given plaquette indices arranged according to
            the underlying shape of the template.

        """
        indices_map = self.get_indices_map_for_instantiation(plaquette_indices)

        element_shape = self._element_scalable_shape.to_numpy_shape(k)
        ret = numpy.zeros(self.shape(k).to_numpy_shape(), dtype=numpy.int_)
        for pos, element in self._layout.items():
            imap = indices_map[pos]
            indices = [imap[i] for i in range(1, element.expected_plaquettes_number + 1)]
            element_instantiation = element.instantiate(k, indices)
            shifted_pos = BlockPosition2D(
                pos.x - self._block_origin.x, pos.y - self._block_origin.y
            )
            ret[
                shifted_pos.y * element_shape[0] : (shifted_pos.y + 1) * element_shape[0],
                shifted_pos.x * element_shape[1] : (shifted_pos.x + 1) * element_shape[1],
            ] = element_instantiation
        return ret

    @override
    def instantiation_origin(self, k: int) -> PlaquettePosition2D:
        return self._block_origin.get_top_left_plaquette_position(self.element_shape(k))



from itertools import zip_longest

import numpy
import stim


def shift_qubits(
    circuit: stim.Circuit,
    *shifts: float,
    also_shift_detectors: bool = True,
) -> stim.Circuit:
    """Shift the qubit coordinates of the provided circuit by ``shifts``.

    Args:
        circuit: circuit containing qubits to shift.
        shifts: a list of shifts to apply to each dimension.
        also_shift_detectors: if ``True``, coordinates of ``DETECTOR``
            instructions are also shifted. Note that this might introduce
            additional dimensions to the ``DETECTOR`` instructions if the
            provided ``shifts`` is contains more elements than the original
            number of dimensions in the arguments.

    Raises:
        TQECError: if any ``QUBIT_COORDS`` instruction in the provided
            ``circuit`` has a number of arguments (i.e., dimensions) that is
            different from the provided number of ``shifts``.

    Returns:
        a new ``stim.Circuit`` instance with qubit coordinates shifted by
        ``shifts``.

    """
    ret = stim.Circuit()
    for instr in circuit:
        if isinstance(instr, stim.CircuitRepeatBlock):
            ret.append(
                stim.CircuitRepeatBlock(
                    instr.repeat_count,
                    shift_qubits(
                        instr.body_copy(),
                        *shifts,
                        also_shift_detectors=also_shift_detectors,
                    ),
                )
            )
        elif instr.name == "QUBIT_COORDS" or (also_shift_detectors and instr.name == "DETECTOR"):
            args = instr.gate_args_copy()
            ret.append(
                instr.name,
                instr.targets_copy(),
                [arg + s for arg, s in zip_longest(args, shifts, fillvalue=0)],
            )
        else:
            ret.append(instr)
    return ret


def shift_to_only_positive(
    circuit: stim.Circuit,
    stick_to_origin: bool = True,
    also_shift_detectors: bool = True,
) -> stim.Circuit:
    """Shift the provided circuit so that it only operates on qubits with positive coordinates.

    Args:
        circuit: quantum circuit to shift.
        stick_to_origin: if ``True``, coordinates that are already positive may
            still be shifted so that the minimum coordinate is ``0``.
        also_shift_detectors: if ``True``, coordinates of ``DETECTOR``
            instructions are also shifted. Note that this might introduce
            additional dimensions to the ``DETECTOR`` instructions if the
            provided ``shifts`` is contains more elements than the original
            number of dimensions in the arguments.

    Returns:
        a copy of ``circuit`` with all the qubit coordinates shifted to positive
        values.

    """
    mins, _ = circuit_bounding_box(circuit)
    shifts = [-m if stick_to_origin or m < 0 else 0 for m in mins]
    return shift_qubits(circuit, *shifts, also_shift_detectors=also_shift_detectors)


def circuit_bounding_box(
    circuit: stim.Circuit,
) -> tuple[list[float], list[float]]:
    """Get the bounding box of the qubits in the provided circuit.

    Note:
        This function uses ``stim.Circuit.get_final_qubit_coordinates`` to get
        qubit coordinates. As such, it returns the "final" bounding box. As long
        as the qubit coordinates are not redefined in the provided circuit, the
        returned bounding box is valid through the whole circuit.

    Args:
        circuit: circuit used to compute the bounding box.

    Returns:
        a tuple ``(mins, maxes)`` containing 2 lists containing the minimum
        (resp. maximum) coordinate found in each dimension.
        If the circuit does not contain any ``QUBIT_COORDS`` instruction, the
        returned lists are empty.

    """
    coordinates = numpy.array(list(circuit.get_final_qubit_coordinates().values()))
    return list(numpy.min(coordinates, axis=0)), list(numpy.max(coordinates, axis=0))



"""Defines the :class:`.Schedule` class.

The :class:`.Schedule` class is a thin wrapper around ``list[int]`` to represent a schedule.
"""

from __future__ import annotations

from collections.abc import Iterator, Sequence
from dataclasses import dataclass, field
from typing import ClassVar

from tqec.circuit.schedule.exception import ScheduleError


@dataclass
class Schedule:
    """Thin wrapper around ``list[int]`` to represent a schedule.

    This class ensures that the list of integers provided is a valid schedule by checking that all
    entries are positive integers, that the list is sorted and that it does not contain any
    duplicate.

    """

    schedule: list[int] = field(default_factory=list)
    """List of integers representing the schedule."""

    _INITIAL_SCHEDULE: ClassVar[int] = 0
    """First entry of a default-constructed schedule."""

    def __post_init__(self) -> None:
        Schedule._check_schedule(self.schedule)

    @staticmethod
    def from_offsets(schedule_offsets: Sequence[int]) -> Schedule:
        """Get a valid schedule from offsets.

        This method should be used to avoid any dependency on
        :py:const:`Schedule._INITIAL_SCHEDULE` in user code.

        """
        return Schedule([Schedule._INITIAL_SCHEDULE + s for s in schedule_offsets])

    @staticmethod
    def _check_schedule(schedule: list[int]) -> None:
        # Check that the schedule is sorted and positive
        if schedule and (
            not all(schedule[i] < schedule[i + 1] for i in range(len(schedule) - 1))
            or schedule[0] < Schedule._INITIAL_SCHEDULE
        ):
            raise ScheduleError(
                f"The provided schedule {schedule} is not a sorted list of positive "
                "integers. You should only provide sorted schedules with positive "
                "entries."
            )

    def __len__(self) -> int:
        return len(self.schedule)

    def __getitem__(self, i: int) -> int:
        return self.schedule[i]

    def __iter__(self) -> Iterator[int]:
        return iter(self.schedule)

    def insert(self, i: int, value: int) -> None:
        """Insert an integer to the schedule.

        If inserting the integer results in an invalid schedule, the schedule is
        brought back to its (valid) original state before calling this function
        and a :py:exc:`~.schedule.exception.ScheduleError` is raised.

        Args:
            i: index at which the provided value should be inserted.
            value: value to insert.

        Raises:
            ScheduleError: if the inserted integer makes the schedule
                invalid.

        """
        self.schedule.insert(i, value)
        try:
            Schedule._check_schedule(self.schedule)
        except ScheduleError as e:
            self.schedule.pop(i)
            raise e

    def append(self, value: int) -> None:
        """Append an integer to the schedule.

        If appending the integer results in an invalid schedule, the schedule is
        brought back to its (valid) original state before calling this function
        and a :py:exc:`~.schedule.exception.ScheduleError` is raised.

        Args:
            value: value to insert.

        Raises:
            ScheduleError: if the inserted integer makes the schedule
                invalid.

        """
        self.schedule.append(value)
        try:
            Schedule._check_schedule(self.schedule)
        except ScheduleError as e:
            self.schedule.pop(-1)
            raise e

    def append_schedule(self, schedule: Schedule) -> None:
        """Append a full schedule **after** ``self``.

        Note:
            The provided ``schedule`` is append just after ``self``. If ``self`` is empty, then we
            have ``self == schedule`` at the end of this method. If ``self`` contains at least one
            schedule, all entries of ``schedule`` are offset by the maximum schedule in ``self``
            plus 1.

        """
        starting_index = self.schedule[-1] + 1 if self.schedule else Schedule._INITIAL_SCHEDULE
        # Not using a generator here but explicitly constructing a list because
        # if `schedule == self` a generator would induce an infinite loop.
        self.schedule.extend([starting_index + s for s in schedule.schedule])

    def pop(self, i: int = -1) -> int:
        """Remove and return the integer at index ``i`` in the schedule.

        Args:
            i: index of the integer to remove. Defaults to ``-1``, i.e., the last
                integer in the schedule.

        Returns:
            the removed integer.

        """
        return self.schedule.pop(i)

    @property
    def max_schedule(self) -> int:
        """Get the maximum timestep in ``self`` or ``0`` if ``self`` is empty."""
        return self.schedule[-1] if self.schedule else 0



from __future__ import annotations

from dataclasses import dataclass, field

from typing_extensions import override

from tqec.circuit.measurement_map import MeasurementRecordsMap
from tqec.compile.blocks.layers.atomic.layout import LayoutLayer
from tqec.compile.detectors.compute import compute_detectors_for_fixed_radius
from tqec.compile.detectors.database import DetectorDatabase
from tqec.compile.tree.annotations import DetectorAnnotation
from tqec.compile.tree.node import LayerNode, NodeWalker
from tqec.plaquette.plaquette import Plaquettes
from tqec.templates.base import Template
from tqec.utils.exceptions import TQECError


@dataclass(frozen=True)
class LookbackInformation:
    """Stores data on one QEC round.

    This data-structure is used to store information that might be useful to
    compute detectors. It only represents one QEC round.

    Attributes:
        template: template representing the QEC round.
        plaquettes: plaquettes that can be used in conjunction with
            ``self.template`` to generate the quantum circuit representing the
            QEC round.
        measurement_records: all the measurement records of the QEC round
            represented by ``self``. This could in theory be computed from
            ``self.template`` and ``self.plaquettes`` by generating the quantum
            circuit and then extracting the measurement records from it, but it
            turns out that we already have access to these records when creating
            such a structure, so we store them to avoid re-computing.

    """

    template: Template
    plaquettes: Plaquettes
    measurement_records: MeasurementRecordsMap


@dataclass
class LookbackInformationList:
    """A sequence of :class:`LookbackInformation` instances."""

    infos: list[LookbackInformation] = field(default_factory=list)

    def append(
        self,
        template: Template,
        plaquettes: Plaquettes,
        measurement_records: MeasurementRecordsMap,
    ) -> None:
        """Add the provided parameters to the lookback window.

        This method might remove older items that should not be considered anymore from the lookback
        stack.
        """
        self.infos.append(LookbackInformation(template, plaquettes, measurement_records))

    def extend(self, other: LookbackInformationList, repetitions: int = 1) -> None:
        """Add the provided lookback information to self, potentially repeating it several times.

        This method can be used when exiting a REPEAT block to update the lookback information by
        taking into account that it might be repeated several times.

        """
        self.infos.extend(other.infos * repetitions)

    def __len__(self) -> int:
        return len(self.infos)

    def __getitem__(self, index: int | slice) -> LookbackInformation | list[LookbackInformation]:
        return self.infos[index]  # pragma: no cover


class LookbackStack:
    def __init__(self) -> None:
        """Initialise the lookback stack.

        The lookback stack can be used to query the current state for detector computation.

        This data-structure keeps information about the past QEC rounds in order to be able to query
        them and help in detector computation by only considering the ``N`` last rounds.

        In particular, this data-structure is useful to keep track of previous rounds in the
        presence of ``REPEAT`` blocks.

        """
        self._stack: list[LookbackInformationList] = [LookbackInformationList()]

    def enter_repeat_block(self) -> None:
        """Append a new entry to the stack."""
        self._stack.append(LookbackInformationList())

    def close_repeat_block(self, repetitions: int) -> None:
        """Remove the last entry on the stack, repeating it as needed into the new last entry."""
        if len(self._stack) < 2:
            raise TQECError(
                f"Only got {len(self._stack)} < 2 entries in the stack. That "
                "means that we are not in a REPEAT block. Cannot call "
                "close_repeat_block()."
            )
        self._stack[-2].extend(self._stack[-1], repetitions)
        self._stack.pop(-1)

    def append(
        self,
        template: Template,
        plaquettes: Plaquettes,
        measurement_records: MeasurementRecordsMap,
    ) -> None:
        """Append a new QEC round in the data-structure."""
        self._stack[-1].append(template, plaquettes, measurement_records)

    def _get_last_n(
        self, n: int
    ) -> tuple[list[Template], list[Plaquettes], list[MeasurementRecordsMap]]:
        if n < 0:
            raise TQECError(
                f"Cannot look back a negative number of rounds. Got a lookback value of {n}."
            )
        if n == 0:
            return [], [], []
        templates: list[Template] = []
        plaquettes: list[Plaquettes] = []
        measurement_records: list[MeasurementRecordsMap] = []
        # Filling the lists in reverse order (i.e., from earlier time to oldest
        # time) and correcting when returning.
        for element in reversed(self._stack):
            for info in reversed(element.infos):
                templates.append(info.template)
                plaquettes.append(info.plaquettes)
                measurement_records.append(info.measurement_records)
                if len(templates) == n:
                    return templates[::-1], plaquettes[::-1], measurement_records[::-1]

        return templates[::-1], plaquettes[::-1], measurement_records[::-1]

    def lookback(
        self,
        n: int,
    ) -> tuple[list[Template], list[Plaquettes], MeasurementRecordsMap]:
        """Get the last ``self._lookback`` QEC rounds."""
        templates, plaquettes, measurement_records = self._get_last_n(n)
        measurement_record = MeasurementRecordsMap()
        for mrec in measurement_records:
            measurement_record = measurement_record.with_added_measurements(mrec)
        return templates, plaquettes, measurement_record

    def __len__(self) -> int:
        if len(self._stack) > 1:
            raise TQECError(
                "Cannot get a meaningful stack length when a REPEAT block is in construction."
            )
        return len(self._stack[0])


class AnnotateDetectorsOnLayerNode(NodeWalker):
    def __init__(
        self,
        k: int,
        manhattan_radius: int = 2,
        detector_database: DetectorDatabase | None = None,
        only_use_database: bool = False,
        lookback: int = 2,
        parallel_process_count: int = 1,
    ):
        """Walker computing and annotating detectors on leaf nodes.

        This class keeps track of the ``lookback`` previous leaf nodes seen and
        uses them to automatically compute the detectors at all the leaf nodes
        it encounters.

        Args:
            k: scaling factor.
            manhattan_radius: Parameter for the automatic computation of detectors.
                Should be large enough so that flows cancelling each other to
                form a detector are strictly contained in plaquettes that are at
                most at a distance of ``manhattan_radius`` from the central
                plaquette. Detector computation runtime grows with this parameter,
                so you should try to keep it to its minimum. A value too low might
                produce invalid detectors.
            detector_database: existing database of detectors that is used to
                avoid computing detectors if the database already contains them.
                Default to `None` which result in not using any kind of database
                and unconditionally performing the detector computation.
            only_use_database: if ``True``, only detectors from the database will be
                used. An error will be raised if a situation that is not registered
                in the database is encountered. Default to ``False``.
            lookback: number of QEC rounds to consider to try to find detectors. Including more
                rounds increases computation time.
            parallel_process_count: number of processes to use for parallel processing.
                1 for sequential processing, >1 for parallel processing using
                ``parallel_process_count`` processes, and -1 for using all available
                CPU cores. Default to 1.

        """
        if lookback < 1:
            raise TQECError(
                "Cannot compute detectors without any layer. The `lookback` "
                f"parameter should be >= 1 but got {lookback}."
            )
        self._k = k
        self._manhattan_radius = manhattan_radius
        self._database = detector_database if detector_database is not None else DetectorDatabase()
        self._only_use_database = only_use_database
        self._lookback_size = lookback
        self._lookback_stack = LookbackStack()
        self._parallel_process_count = parallel_process_count

    @override
    def visit_node(self, node: LayerNode) -> None:
        if not isinstance(node._layer, LayoutLayer):
            return
        annotations = node.get_annotations(self._k)
        if annotations.circuit is None:
            raise TQECError("Cannot compute detectors without the circuit annotation.")
        self._lookback_stack.append(
            *node._layer.to_template_and_plaquettes(),
            MeasurementRecordsMap.from_scheduled_circuit(annotations.circuit),
        )
        templates, plaquettes, measurement_records = self._lookback_stack.lookback(
            self._lookback_size
        )

        detectors = compute_detectors_for_fixed_radius(
            templates,
            self._k,
            plaquettes,
            self._manhattan_radius,
            self._database,
            self._only_use_database,
            self._parallel_process_count,
        )

        for detector in detectors:
            annotations.detectors.append(
                DetectorAnnotation.from_detector(detector, measurement_records)
            )

    @override
    def enter_node(self, node: LayerNode) -> None:
        if node.is_repeated:
            self._lookback_stack.enter_repeat_block()

    @override
    def exit_node(self, node: LayerNode) -> None:
        if not node.is_repeated:
            return
        # Note: this is the place to perform checks. In particular, checking that
        # detectors computed at the first repetition of the REPEAT block are also
        # valid at any repetitions. This is a requirement for the REPEAT block to
        # make sense, but that would be nice to include a check to avoid
        # misleadingly include detectors that are incorrect sometimes.
        repetitions = node.repetitions
        assert repetitions is not None
        self._lookback_stack.close_repeat_block(repetitions.integer_eval(self._k))



"""Defines several helper methods to facilitate QEC circuit simulation."""

from .plotting import plot_observable_as_inset as plot_observable_as_inset
from .simulation import start_simulation_using_sinter as start_simulation_using_sinter
from .split import heuristic_custom_error_key as heuristic_custom_error_key
from .split import split_stats_for_observables as split_stats_for_observables



from __future__ import annotations

from collections.abc import Iterable, Mapping, Sequence
from itertools import chain

from typing_extensions import override

from tqec.compile.blocks.enums import SpatialBlockBorder, TemporalBlockBorder
from tqec.compile.blocks.layers.atomic.base import BaseLayer
from tqec.compile.blocks.layers.atomic.layout import LayoutLayer
from tqec.compile.blocks.layers.composed.base import BaseComposedLayer
from tqec.utils.exceptions import TQECError
from tqec.utils.scale import LinearFunction, PhysicalQubitScalable2D


class SequencedLayers(BaseComposedLayer):
    def __init__(
        self,
        layer_sequence: Sequence[BaseLayer | BaseComposedLayer],
        trimmed_spatial_borders: frozenset[SpatialBlockBorder] = frozenset(),
    ):
        """Composed layer implementing a fixed sequence of layers.

        This composed layer sequentially applies layers from a fixed sequence.

        Args:
            layer_sequence: non-empty sequence of layers to apply one after the
                other. All the layers in this sequence should have exactly the same
                spatial footprint.
            trimmed_spatial_borders: all the spatial borders that have been
                removed from the layer.

        Raises:
            TQECError: if the provided ``layer_sequence`` is empty.

        """
        super().__init__(trimmed_spatial_borders)
        self._layer_sequence = layer_sequence
        self._post_init_check()

    @property
    def layer_sequence(self) -> Sequence[BaseLayer | BaseComposedLayer]:
        """Get the sequence of layers stored by ``self``."""
        return self._layer_sequence

    def _post_init_check(self) -> None:
        if len(self.layer_sequence) < 1:
            raise TQECError(
                f"An instance of {type(self).__name__} is expected to have "
                f"at least one layer. Found {len(self.layer_sequence)}."
            )

    @property
    def schedule(self) -> tuple[LinearFunction, ...]:
        """Return the duration of each of the sequenced layers."""
        return tuple(layer.scalable_timesteps for layer in self.layer_sequence)

    @property
    @override
    def scalable_timesteps(self) -> LinearFunction:
        return sum(
            (layer.scalable_timesteps for layer in self.layer_sequence),
            start=LinearFunction(0, 0),
        )

    @property
    @override
    def scalable_shape(self) -> PhysicalQubitScalable2D:
        if any(isinstance(layer, LayoutLayer) for layer in self._layer_sequence):
            raise NotImplementedError(
                f"Computation of the scalable_shape for {LayoutLayer.__name__} "
                "instances has not been implemented yet."
            )
        scalable_shape = self._layer_sequence[0].scalable_shape
        for layer in self._layer_sequence[1:]:
            if layer.scalable_shape != scalable_shape:
                raise TQECError("Found a different scalable shape.")
        return scalable_shape

    def _layers_with_spatial_borders_trimmed(
        self, borders: Iterable[SpatialBlockBorder]
    ) -> list[BaseLayer | BaseComposedLayer]:
        return [layer.with_spatial_borders_trimmed(borders) for layer in self.layer_sequence]

    @override
    def with_spatial_borders_trimmed(
        self, borders: Iterable[SpatialBlockBorder]
    ) -> SequencedLayers:
        return SequencedLayers(
            self._layers_with_spatial_borders_trimmed(borders),
            self.trimmed_spatial_borders | frozenset(borders),
        )

    def _layers_with_temporal_borders_replaced(
        self, border_replacements: Mapping[TemporalBlockBorder, BaseLayer | None]
    ) -> list[BaseLayer | BaseComposedLayer]:
        layers = list(self.layer_sequence)
        if (border := TemporalBlockBorder.Z_NEGATIVE) in border_replacements:
            first_layer = layers[0].with_temporal_borders_replaced(
                {border: border_replacements[border]}
            )
            if first_layer is not None:
                layers[0] = first_layer
            else:
                layers.pop(0)
        if (border := TemporalBlockBorder.Z_POSITIVE) in border_replacements:
            last_layer = layers[-1].with_temporal_borders_replaced(
                {border: border_replacements[border]}
            )
            if last_layer is not None:
                layers[-1] = last_layer
            else:
                layers.pop(-1)
        return layers

    @override
    def with_temporal_borders_replaced(
        self, border_replacements: Mapping[TemporalBlockBorder, BaseLayer | None]
    ) -> BaseLayer | BaseComposedLayer | None:
        if not border_replacements:
            return self
        layers = self._layers_with_temporal_borders_replaced(border_replacements)
        match len(layers):
            case 0:
                return None
            case 1:
                return layers[0]
            case _:
                return SequencedLayers(layers)

    @override
    def all_layers(self, k: int) -> Iterable[BaseLayer]:
        yield from chain.from_iterable(
            ((layer,) if isinstance(layer, BaseLayer) else layer.all_layers(k))
            for layer in self.layer_sequence
        )

    @override
    def to_sequenced_layer_with_schedule(
        self, schedule: tuple[LinearFunction, ...]
    ) -> SequencedLayers:
        duration = sum(schedule, start=LinearFunction(0, 0))
        if self.scalable_timesteps != duration:
            raise TQECError(
                f"Cannot transform the {SequencedLayers.__name__} instance to a "
                f"{SequencedLayers.__name__} instance with the provided schedule. "
                f"The provided schedule has a duration of {duration} but the "
                f"instance to transform has a duration of {self.scalable_timesteps}."
            )
        if self.schedule == schedule:
            return self
        raise NotImplementedError(
            f"Adapting a {SequencedLayers.__name__} instance to another "
            "schedule is not yet implemented."
        )

    def __eq__(self, value: object) -> bool:
        return isinstance(value, SequencedLayers) and self.layer_sequence == value.layer_sequence

    def __hash__(self) -> int:
        raise NotImplementedError(f"Cannot hash efficiently a {type(self).__name__}.")

    @override
    def get_temporal_layer_on_border(self, border: TemporalBlockBorder) -> BaseLayer:
        return self._layer_sequence[
            0 if border == TemporalBlockBorder.Z_NEGATIVE else -1
        ].get_temporal_layer_on_border(border)

    @property
    @override
    def scalable_num_moments(self) -> LinearFunction:
        return sum(
            (layer.scalable_num_moments for layer in self.layer_sequence),
            start=LinearFunction(0, 0),
        )



"""Defines enumerations that are only used in :mod:`tqec.templates`."""

from __future__ import annotations

from enum import Enum, auto


class TemplateBorder(Enum):
    TOP = auto()
    BOTTOM = auto()
    LEFT = auto()
    RIGHT = auto()

    def opposite(self) -> TemplateBorder:  # pragma: no cover
        """Return the opposite border."""
        match self:
            case TemplateBorder.TOP:
                return TemplateBorder.BOTTOM
            case TemplateBorder.BOTTOM:
                return TemplateBorder.TOP
            case TemplateBorder.LEFT:
                return TemplateBorder.RIGHT
            case TemplateBorder.RIGHT:
                return TemplateBorder.LEFT
            # wildcard entry added as it was flagged by ty
            case _:
                raise ValueError(f"Cannot return the opposite border of {self}.")



import json
from collections.abc import Sequence
from multiprocessing import Pool, cpu_count

import numpy
import numpy.typing as npt
import stim
from tqecd.flow import build_flows_from_fragments
from tqecd.fragment import Fragment
from tqecd.match import (
    MatchedDetector,
    match_boundary_stabilizers,
    match_detectors_within_fragment,
)

from tqec.circuit.measurement import Measurement, get_measurements_from_circuit
from tqec.circuit.qubit import GridQubit
from tqec.circuit.schedule import ScheduledCircuit, relabel_circuits_qubit_indices
from tqec.compile.detectors.database import DetectorDatabase
from tqec.compile.detectors.detector import Detector
from tqec.compile.generation import generate_circuit_from_instantiation
from tqec.plaquette.plaquette import Plaquettes
from tqec.templates.base import Template
from tqec.templates.display import get_template_representation_from_instantiation
from tqec.templates.subtemplates import (
    SubTemplateType,
    get_spatially_distinct_3d_subtemplates,
)
from tqec.utils.array import to2dlist
from tqec.utils.coordinates import StimCoordinates
from tqec.utils.exceptions import TQECError
from tqec.utils.position import PlaquettePosition2D, Shift2D


def _get_measurement_offset_mapping(circuit: stim.Circuit) -> dict[int, Measurement]:
    """Get a mapping from measurement offsets to a :class:`.Measurement` instance.

    The measurement offsets used as keys in the returned mapping are in accordance with the
    ``tqecd.detectors`` external package.

    This function returns the mapping from negative offsets that are
    supposed to each represent a unique measurement in the circuit to
    `Measurement` instances, that serve the same purpose but use a different
    encoding.

    Note:
        As a sanity check, the user of this function is encouraged to check
        if the qubits in each `Measurement` instances returned as values of
        the mapping corresponds to the expected qubit.

    Args:
        circuit: the circuit to create the mapping for.

    Returns:
        a mapping from record offset to :class:`Measurement` for all the
        measurements appearing in the provided `circuit`.

    """
    return {-i - 1: m for i, m in enumerate(reversed(get_measurements_from_circuit(circuit)))}


def _matched_detectors_to_detectors(
    detectors: list[MatchedDetector],
    measurements_by_offset: dict[int, Measurement],
) -> list[Detector]:
    """Transform :class:`MatchedDetector` instances into `Detector` ones.

    Args:
        detectors: list of detectors to translate.
        measurements_by_offset: map from measurement record offsets that are used
            in the provided `detectors` to :class:`Measurement` instances
            using a spatially-local representation of measurements.

    Returns:
        :class:`Detector` instances representing the same detectors as the
        provided `detectors`.

    """
    ret: list[Detector] = []
    for d in detectors:
        measurements: list[Measurement] = [measurements_by_offset[m.offset] for m in d.measurements]
        x, y, t = d.coords
        ret.append(Detector(frozenset(measurements), StimCoordinates(x, y, t)))
    return ret


def _center_plaquette_syndrome_qubits(
    subtemplate: SubTemplateType, plaquettes: Plaquettes, increments: Shift2D
) -> list[GridQubit]:
    """Return the syndrome qubits used by the central plaquette of the provided ``subtemplate``.

    The qubits are returned in the sub-template coordinates (i.e., origin at top-left corner of the
    provided `subtemplate`).

    Note:
        This function only returns a subset of the syndrome qubits. This is
        because, for some plaquettes (e.g., extended stabilizer measurement),
        some corner qubits are syndrome qubits, but are shared by 4 plaquettes.
        For this reason, each plaquette "owns" its top-left qubit. Following
        this rule, there are a few cases where a data-qubit is owned by nobody,
        hence a few exceptions to the above rule are added.

    Warning:
        If the provided ``subtemplate`` has been obtained from a manhattan radius
        ``r == 0`` (i.e., no neighbouring plaquette is taken into account), only
        the top-left data-qubit is owned by the plaquette, and no exception come
        into play.

    Args:
        subtemplate: 2-dimensional array representing the sub-template we are
            interested in.
        plaquettes: a collection of plaquettes that will be used to generate a
            circuit from the provided `subtemplate`.
        increments: spatial increments between each `Plaquette` origin.

    Returns:
        a collection of qubits that are used as syndrome qubits by the central
        plaquette. Returns an empty collection if the central plaquette has the
        index `0`.

    """
    # Subtemplates are expected to have a shape of (2*r+1, 2*r+1), so `r` can
    # be recovered by computing `(2*r+1) // 2 == r`.
    r = subtemplate.shape[0] // 2
    central_plaquette_index = int(subtemplate[r, r])

    if central_plaquette_index == 0:
        return []

    # In the following, "X" is a qubit that "belongs" to the plaquette drawn
    # using "=" and "|", "O" is a qubit that does not, and plaquette(s) drawn
    # using "~" and "'" are representing empty plaquettes. Numbers in the center
    # of empty plaquettes correspond to the case applied to justify the absence
    # of owner.

    # Case 1, General case, always valid:
    # X ===== O
    # |       |
    # |   X   |
    # |       |
    # O ===== O
    considered_syndrome_qubits = {GridQubit(0, 0), GridQubit(-1, -1)}
    # Case 2, when the top-right qubit should be added because no plaquette on
    # the left to own it.
    # X ===== X ~~~~~ O
    # |       |       '
    # |   X   |   1   '
    # |       |       '
    # O ===== O ~~~~~ O
    if r != 0 and subtemplate[r, r + 1] == 0:
        considered_syndrome_qubits |= {GridQubit(1, -1)}
    # When the bottom-left qubit should be added because no plaquette on the
    # bottom and bottom-left to own it.
    #         X ===== O
    #         |       |
    #         |   X   |
    #         |       |
    # O ~~~~~ X ===== O
    # '       '       '
    # '   2   '   1   '
    # '       '       '
    # O ~~~~~ O ~~~~~ O
    if r != 0 and (subtemplate[r + 1, r - 1] == subtemplate[r + 1, r] == 0):
        considered_syndrome_qubits |= {GridQubit(-1, 1)}
    # When the bottom-right qubit should be added because no plaquette on the
    # bottom, bottom-right and right to own it.
    # X ===== O ~~~~~ O
    # |       |       '
    # |   X   |   3   '
    # |       |       '
    # O ===== X ~~~~~ O
    # '       '       '
    # '   2   '   1   '
    # '       '       '
    # O ~~~~~ O ~~~~~ O
    if r != 0 and (
        subtemplate[r + 1, r - 1] == subtemplate[r + 1, r] == subtemplate[r, r + 1] == 0
    ):
        considered_syndrome_qubits |= {GridQubit(1, 1)}

    central_plaquette = plaquettes[central_plaquette_index]
    origin = central_plaquette.origin
    offset = Shift2D(r * increments.x + origin.x, r * increments.y + origin.y)
    return [
        q + offset
        for q in central_plaquette.qubits.syndrome_qubits
        if q in considered_syndrome_qubits
    ]


def _best_effort_filter_detectors(
    detectors: list[Detector],
    subtemplates: Sequence[SubTemplateType],
    plaquettes: Sequence[Plaquettes],
    increments: Shift2D,
) -> frozenset[Detector]:
    """Filter detectors using a best-effort strategy.

    This function filters out detectors that do not involve at least one measurement on a syndrome
    qubit of the central plaquette in the last round. Such a filtering is voluntarily not too
    strict, as the goal is to reduce the number of detectors that should be considered, but a more
    robust filter will be applied later in the pipeline. So this function implements a good-enough
    filtering that is not perfect, but that is at least guaranteed not to remove detectors that
    should not be removed.

    Warning:
        This function assumes that there is exactly one measurement on each
        syndrome qubit per round. That means that if a syndrome qubit is not
        measured, or if it is measured twice during the same round, this function
        will return wrong results.
        For the moment, this assumption is verified for all the plaquettes we are
        using, but this restriction should be kept in mind in case a future
        plaquette does not check this condition.

    Warning:
        This function tries as much as possible to filter the maximum number of
        detectors for the provided ``subtemplates`` and ``plaquettes``. For the
        moment, this filtering is not perfect and detectors might end up
        duplicated on several combos of ``subtemplates`` and ``plaquettes``.

        That is not a problem as long as the number of duplicated detectors is
        relatively low and a second more robust filter based on deduplication
        via ``set`` is in place.

    Args:
        detectors: list of detectors to filter.
        subtemplates: subtemplate on which the detectors have been computed. Only
            the last time step (`subtemplates[-1]`) is used by this function.
        plaquettes: plaquettes on which the detectors have been computed. Only
            the last time step (`plaquettes[-1]`) is used by this function.
        increments: increment between two neighbouring plaquette origins.

    Returns:
        a filtered sequence of detectors. All the returned detectors are from
        the provided `detectors` (i.e., this function does not create or combine
        detectors).

    """
    # First, we want the detectors to be composed of at least one measurement
    # involving one of the syndrome qubits of the central plaquette in the last
    # timestep.
    central_syndrome_qubits_measurements = frozenset(
        # Assumes that syndrome qubits are only measured once per round, which
        # seems reasonable. But still, this is an important assumption.
        Measurement(q, -1)
        for q in _center_plaquette_syndrome_qubits(subtemplates[-1], plaquettes[-1], increments)
    )
    filtered_detectors = [
        d for d in detectors if d.measurements.intersection(central_syndrome_qubits_measurements)
    ]
    return frozenset(filtered_detectors)


def _compute_detectors_at_end_of_situation(
    subtemplates: Sequence[SubTemplateType],
    plaquettes: Sequence[Plaquettes],
    increments: Shift2D,
) -> frozenset[Detector]:
    if len(plaquettes) != len(subtemplates):
        raise TQECError(
            "Unsupported input: you should provide as many subtemplates as there are plaquettes."
        )
    # Note: if the center plaquette of the last entry of `subtemplates` does not
    #       contain any measurement, then we can be sure that there is no
    #       detectors here, so early return.
    radius = subtemplates[-1].shape[0] // 2
    center_plaquette_index = subtemplates[-1][radius, radius]
    if center_plaquette_index == 0:
        return frozenset()
    center_plaquette = plaquettes[-1][center_plaquette_index]
    if center_plaquette.num_measurements == 0:
        return frozenset()

    # Note: if there is more than 1 time slice, remove any initial time slice
    # that is trivially empty. This is a faster version of the next while loop,
    # but this might not catch all empty circuits (e.g., those that have an
    # explicit plaquette, that turns out to be empty).
    while len(subtemplates) > 1 and numpy.all(subtemplates[0] == 0):
        assert len(plaquettes) > 1  # make type checkers happy
        subtemplates = subtemplates[1:]
        plaquettes = plaquettes[1:]

    # Note: if there is more than 1 time slice, remove any initial time slice
    # that is empty. Same as above, but without any false negative and less
    # efficient.
    current_circuit = generate_circuit_from_instantiation(
        subtemplates[0], plaquettes[0], increments
    )
    while len(subtemplates) > 1 and current_circuit.is_empty():
        assert len(plaquettes) > 1  # make type checkers happy
        subtemplates = subtemplates[1:]
        plaquettes = plaquettes[1:]
        current_circuit = generate_circuit_from_instantiation(
            subtemplates[0], plaquettes[0], increments
        )
    # Build subcircuit for each Plaquettes layer
    subcircuits: list[ScheduledCircuit] = [current_circuit]
    for subtemplate, plaqs in zip(subtemplates[1:], plaquettes[1:]):
        subcircuit = generate_circuit_from_instantiation(subtemplate, plaqs, increments)
        subcircuits.append(subcircuit)
    # Extract the global qubit map from the generated sub-circuits, relabeling
    # qubits if needed.
    subcircuits, global_qubit_map = relabel_circuits_qubit_indices(subcircuits)
    # Get the stim.Circuit instances. We do not need the coordinates here
    # because we have already computed the global qubit map.
    coordless_subcircuits = [sc.get_circuit(include_qubit_coords=False) for sc in subcircuits]
    # Get the full stim.Circuit to compute a measurement records offset map and
    # filter out detectors at the end.
    complete_circuit = global_qubit_map.to_circuit()
    for coordless_subcircuit in coordless_subcircuits[:-1]:
        complete_circuit += coordless_subcircuit
        complete_circuit.append("TICK", [], [])
    complete_circuit += coordless_subcircuits[-1]

    # Use tqecd.detectors module to match the detectors. Note that, for
    # the moment, only the last two time slices are taken into account.
    coordinates_by_index = {i: (float(q.x), float(q.y)) for i, q in global_qubit_map.items()}
    fragments = [Fragment(circ) for circ in coordless_subcircuits]
    flows = build_flows_from_fragments(fragments)
    matched_detectors = match_detectors_within_fragment(flows[-1], coordinates_by_index)
    if len(flows) == 2:
        matched_detectors.extend(
            match_boundary_stabilizers(flows[-2], flows[-1], coordinates_by_index)
        )
    # Note that the matched detectors do not have a time coordinate yet. Because
    # all the matched detectors belong to the last flow, the time coordinate can
    # just be "0". Simply add that to all detectors.
    matched_detectors = [d.with_time_coordinate(0) for d in matched_detectors]

    # Get the detectors as Detector instances instead of the
    # `tqecd.MatchedDetector` class.
    measurements_by_offset = _get_measurement_offset_mapping(complete_circuit)
    detectors = _matched_detectors_to_detectors(matched_detectors, measurements_by_offset)

    # Filter out detectors and return the left ones.
    return _best_effort_filter_detectors(detectors, subtemplates, plaquettes, increments)


def _get_database_access_exception(
    subtemplates: Sequence[SubTemplateType],
    plaquettes_by_timestep: Sequence[Plaquettes],
) -> TQECError:
    return TQECError(
        "Failed to retrieve a situation from the provided database but "
        "only_use_database was True. Failing instead of computing "
        "automatically detectors. You might want to populate the "
        "database with the missing situation before re-calling this "
        "method. Description of the situation:\n"
        "Subtemplates and plaquettes (by decreasing time, the first "
        "displayed subtemplate is for the last timeslice):\n"
        + ("\n" + "-" * 40 + "\n").join(
            "\n".join(
                (
                    "Subtemplate:",
                    get_template_representation_from_instantiation(subtemplate),
                    "Plaquettes:",
                    json.dumps(plaquettes.to_name_dict()),
                )
            )
            for subtemplate, plaquettes in zip(subtemplates, plaquettes_by_timestep)
        )
    )


def compute_detectors_at_end_of_situation(
    subtemplates: Sequence[SubTemplateType],
    plaquettes_by_timestep: Sequence[Plaquettes],
    increments: Shift2D,
    database: DetectorDatabase | None = None,
    only_use_database: bool = False,
    parallel_process_count: int = 1,
) -> frozenset[Detector]:
    """Return detectors that should be added at the end of the provided situation.

    Args:
        subtemplates: a sequence of sub-template(s), each entry consisting of
            a square 2-dimensional array of integers with odd-length sides
            representing the arrangement of plaquettes in a subtemplate.
        plaquettes_by_timestep: a sequence of collection of plaquettes each
            representing one QEC round.
        increments: spatial increments between each `Plaquette` origin.
        database: existing database of detectors that is used to avoid computing
            detectors if the database already contains them. If provided, this
            function guarantees that the database will contain the provided
            situation when returning (i.e., either it already contained the
            situation or it has been updated **in-place** with the computed
            detectors). If the database is frozen and a new situation is
            encountered, an exception will be thrown when trying to mutate the
            database. Default to `None` which result in not using any kind of
            database and unconditionally performing the detector computation.
        only_use_database: if ``True``, only detectors from the database will be
            used. An error will be raised if a situation that is not registered
            in the database is encountered or if the database is not provided.
            Default to ``False``.
        parallel_process_count: number of processes to use for parallel processing.
            1 for sequential processing, >1 for parallel processing using
            ``parallel_process_count`` processes, and -1 for using all available
            CPU cores. Default to 1.

    Returns:
        all the detectors that can be appended at the end of the circuit
        represented by the provided `subtemplates` and `plaquettes_at_timestep`.

    Raises:
        TQECError: if `len(subtemplates) != len(plaquettes_at_timestep)`.

    """
    # Try to recover the result from the database.
    if database is not None:
        detectors = database.get_detectors(subtemplates, plaquettes_by_timestep)
        # If not found and only detectors from the database should be used, this
        # is an error.
        if detectors is None and only_use_database:
            raise _get_database_access_exception(subtemplates, plaquettes_by_timestep)
        # Else, if not found but we are allowed to compute detectors, compute
        # and store in database.
        elif detectors is None:
            detectors = _compute_detectors_at_end_of_situation(
                subtemplates, plaquettes_by_timestep, increments
            )
            database.add_situation(subtemplates, plaquettes_by_timestep, detectors)
    # If database is None
    else:
        if only_use_database:
            raise _get_database_access_exception(subtemplates, plaquettes_by_timestep)
        detectors = _compute_detectors_at_end_of_situation(
            subtemplates, plaquettes_by_timestep, increments
        )

    # If parallel processing is not enabled, shift the detectors here.
    # Otherwise, wait until all child processes have finished computation.
    # In that case, update the database with the computed detectors in the parent process,
    # and then shift the detectors afterwards.
    if parallel_process_count == 1:
        detectors = _shift_detectors_to_center_of_subtemplate(detectors, subtemplates, increments)
    return detectors


def _shift_detectors_to_center_of_subtemplate(
    detectors: frozenset[Detector],
    subtemplates: Sequence[SubTemplateType],
    increments: Shift2D,
) -> frozenset[Detector]:
    # `subtemplate.shape` should be `(2 * radius + 1, 2 * radius + 1)` so we can
    # recover the radius with the below expression.
    radius = subtemplates[0].shape[0] // 2

    # We have a coordinate system change to apply to `detectors`.
    # `detectors` is using a coordinate system with the origin at the
    # top-left corner of the current sub-template, but we need to return
    # detectors that use the central plaquette origin as their coordinate system
    # origin.
    shift_x, shift_y = -radius * increments.x, -radius * increments.y

    return frozenset(d.offset_spatially_by(shift_x, shift_y) for d in detectors)


def _get_or_default(
    array: npt.NDArray[numpy.int_], slices: Sequence[tuple[int, int]], default: int = 0
) -> npt.NDArray[numpy.int_]:
    """Get slices of an array, returning the provided ``default`` value for out-of-bound accesses.

    Args:
        array: `numpy` array to recover values from.
        slices: a sequence of tuples `(start, stop)` representing the slice that
            should be returned for the corresponding array axis. The first slice
            indexes elements on `axis=0`, the second on `axis=1`, ... The start
            is inclusive, the stop is exclusive.
        default: value to use when indices from the provided slices are
            out-of-bound for the provided `array`. Defaults to 0.

    Raises:
        TQECError: if any of the provided slice has `start > stop`.
        TQECError: if the number of provided slices does not exactly match
            the number of dimensions of the provided array.

    Returns:
        `array[slices[0][0]:slices[0][1], ..., slices[-1][0]:slices[-1][1]]`,
        with any out-of-bounds index associated to the provided `default` value.

    """
    if any(start > stop for start, stop in slices):
        raise TQECError("The provided slices should be non-empty.")
    if len(slices) != len(array.shape):
        raise TQECError(
            f"Expected one slice per array dimension. Got {len(slices)} slices "
            f"but {len(array.shape)} dimensions to the provided array."
        )
    slice_shape = tuple(stop - start for start, stop in slices)
    ret = numpy.full(slice_shape, fill_value=default, dtype=numpy.int_)

    # Now, fill ret with entries from the provided array. We should make sure that
    # we do not try to access elements from `array` that do not exist (out of
    # bounds).
    ret_slices: list[slice] = []
    array_slices: list[slice] = []
    for (start_slice, stop_slice), array_bound in zip(slices, array.shape):
        start_array, stop_array = 0, array_bound
        # There are 6 different cases:
        #      start_array                                   stop_array
        #           [--------------------------------------------[
        #
        # 1: [--[
        # 2:                                                        [--[
        # 3:    [------------------------------------------------------[
        # 4:    [---------------------[
        # 5:                                           [---------------[
        # 6:                   [--------------------------[

        # In cases 1 or 2, there is nothing to take from the array so append an
        # empty slice.
        if stop_slice <= start_array or start_slice >= stop_array:
            ret_slices.append(slice(0, 0))
            array_slices.append(slice(0, 0))
        # In case 3, the slice covers the array, so take the whole dimension of
        # the array.
        elif start_slice <= start_array and stop_array <= stop_slice:
            ret_slices.append(slice(start_array - start_slice, stop_array - start_slice))
            array_slices.append(slice(start_array, stop_array))
        # In case 4, only a part of the slice covers the array.
        elif start_slice <= start_array < stop_slice <= stop_array:
            ret_slices.append(slice(start_array - start_slice, stop_slice - start_slice))
            array_slices.append(slice(start_array, stop_slice))
        # In case 5, only a part of the slice covers the array.
        elif start_array <= start_slice < stop_array <= stop_slice:
            ret_slices.append(slice(0, stop_array - start_slice))
            array_slices.append(slice(start_slice, stop_array))
        # In case 6, the whole slice is within bounds, to use the whole slice.
        elif start_array <= start_slice <= stop_slice <= stop_array:
            ret_slices.append(slice(0, stop_slice - start_slice))
            array_slices.append(slice(start_slice, stop_slice))
        else:
            raise NotImplementedError(
                f"Trying to get the slice [{start_slice}, {stop_slice}[ from an "
                f"array of size {stop_array}. The case should be covered, but "
                "none of the conditions matched, which hints at a mistake "
                "somewhere."
            )

    ret[tuple(ret_slices)] = array[tuple(array_slices)]
    return ret


def _compute_superimposed_template_instantiations(
    templates: Sequence[Template], k: int
) -> list[npt.NDArray[numpy.int_]]:
    """Compute the instantiation of all the provided ``templates``, taking care of alignment.

    When instantiating multiple templates that are supposed to be stacked up on
    top of each other (i.e., executed one after the other) we might be
    confronted with several edge cases:

    - what if the templates do not have the same shape?
    - what if the templates do not have the same origin?

    In our specific case of computing detectors, we always focus on the top-most
    (i.e., last executed, last entry of the provided `templates`) template because
    this is the template that we are searching detectors in. That means that we
    want to instantiate :class:`Template` instances with potentially different
    shapes and origins and be sure that they are all aligned with the top-most
    :class:`Template` instance.

    This function ensures exactly this. It does that by instantiating all the
    provided templates and cutting all the obtained instantiations to the
    coordinates where the last provided template is defined.

    Args:
        templates: instances representing templates that will be instantiated
            and cut to the coordinates where `templates[-1]` is defined.
        k: scaling parameter used to instantiate `templates`.

    Returns:
        a list of template instantiation that all have the same shape and the
        same origin, meaning that in the final circuit `ret[i][x, y]` represent
        indices of plaquettes that will be stacked up (in time) for all `i` and
        any `x` and `y`.

    """
    origins = [t.instantiation_origin(k) for t in templates]
    instantiations = [t.instantiate(k) for t in templates]

    top_left = origins[-1]
    n, m = instantiations[-1].shape
    bottom_right = PlaquettePosition2D(top_left.x + m, top_left.y + n)

    # Get the correct instantiations
    ret: list[npt.NDArray[numpy.int_]] = []
    for inst_origin, instantiation in zip(origins, instantiations):
        ret.append(
            _get_or_default(
                instantiation,
                slices=[
                    (top_left.y - inst_origin.y, bottom_right.y - inst_origin.y),
                    (top_left.x - inst_origin.x, bottom_right.x - inst_origin.x),
                ],
                default=0,
            )
        )
    return ret


def _extract_subtemplates_from_s3d(
    s3d: npt.NDArray[numpy.int_],
) -> list[npt.NDArray[numpy.int_]]:
    """Extract 2D spatial subtemplates from a 3D array.

    This function takes a 3D array where the first two dimensions
    represent spatial coordinates and the third dimension represents time steps.
    It extracts a list of 2D spatial arrays, one for each time step.

    Args:
        s3d: 3D numpy array representation of sub-templates.

    Returns:
        List of 2D numpy arrays, one for each time step.

    """
    return [s3d[:, :, i] for i in range(s3d.shape[2])]


def _compute_detector_for_subtemplate(
    args: tuple[
        tuple[int, ...],  # indices
        npt.NDArray[numpy.int_],  # s3d
        Sequence[Plaquettes],  # plaquettes
        Shift2D,  # increments
        int,  # parallel_process_count
    ],
) -> tuple[tuple[int, ...], frozenset[Detector]]:
    """Wrap :func:`compute_detectors_at_end_of_situation` for parallel processing of detectors.

    Args:
        args: A tuple containing:
            - indices: Tuple of indices identifying the subtemplate
            - s3d: 3D numpy array representing the subtemplate
            - plaquettes: Sequence of plaquettes for each time slice
            - increments: Spatial increments between plaquette origins
            - only_use_database: Whether to only use the database

    Returns:
        A tuple containing the indices and the computed detectors

    """
    indices, s3d, plaquettes, increments, parallel_process_count = args
    return (
        indices,
        compute_detectors_at_end_of_situation(
            _extract_subtemplates_from_s3d(s3d),
            plaquettes,
            increments,
            # Currently, we do not find an efficient way to share the database between
            # multiple processes, so we just pass `None` here.
            database=None,
            only_use_database=False,
            parallel_process_count=parallel_process_count,
        ),
    )


def compute_detectors_for_fixed_radius(
    templates: Sequence[Template],
    k: int,
    plaquettes: Sequence[Plaquettes],
    fixed_subtemplate_radius: int = 2,
    database: DetectorDatabase | None = None,
    only_use_database: bool = False,
    parallel_process_count: int = 1,
) -> list[Detector]:
    """Compute and returns detectors from the provided computation description.

    This function computes the detectors that should be added at the end of the circuit that would
    be obtained from the provided ``template_at_timestep`` and ``plaquettes_at_timestep`` and
    returns them.

    Using a template + plaquette approach allows for efficient detector computation.

    Args:
        templates: a sequence containing `t` :class:`Template` instance(s), each
            representing one QEC round.
        k: scaling factor to consider in order to instantiate the provided
            template.
        plaquettes: a sequence containing `t` collection(s) of plaquettes each
            representing one QEC round.
        fixed_subtemplate_radius: Manhattan radius to consider when splitting the
            provided `template` into sub-templates. Should be large enough so
            that flows cancelling each other to form a detector are strictly
            contained in the sub-template and cannot escape from it (which is
            mostly equivalent to say that flows should not interact with qubits
            on the border of the sub-templates).
        database: existing database of detectors that is used to avoid computing
            detectors if the database already contains them. If provided, this
            function guarantees that the database will contain the provided
            situation when returning (i.e., either it already contained the
            situation or it has been updated **in-place** with the computed
            detectors). Default to `None` which result in not using any kind of
            database and unconditionally performing the detector computation.
        only_use_database: if ``True``, only detectors from the database will be
            used. An error will be raised if a situation that is not registered
            in the database is encountered. Default to ``False``.
        parallel_process_count: number of processes to use for parallel processing.
            1 for sequential processing, >1 for parallel processing using
            ``parallel_process_count`` processes, and -1 for using all available
            CPU cores. Default to 1.

    Returns:
        a collection of detectors that should be added at the end of the circuit
        that would be obtained from the provided `templates` and `plaquettes`.

    """
    all_increments = frozenset(t.get_increments() for t in templates)
    if len(all_increments) != 1:
        raise TQECError(
            "Expected all the provided templates to have the same increments. "
            f"Found the following different increments: {all_increments}."
        )
    increments = next(iter(all_increments))

    if len(templates) != len(plaquettes):
        raise TQECError("Expecting the same number of entries in templates and plaquettes.")

    template_instantiations = _compute_superimposed_template_instantiations(templates, k)
    unique_3d_subtemplates = get_spatially_distinct_3d_subtemplates(
        template_instantiations, manhattan_radius=fixed_subtemplate_radius
    )

    # Each detector in detectors_by_subtemplate is using a coordinate system
    # centered on the central plaquette origin.
    detectors_by_subtemplate: dict[tuple[int, ...], frozenset[Detector]] = {}

    # Handle the special case of parallel_process_count == -1
    if parallel_process_count == -1:
        parallel_process_count = cpu_count()
    # If parallel_process_count > 1 we will enable parallel processing to
    # compute detectors in parallel.
    if parallel_process_count > 1:
        args_list = [
            (indices, s3d, plaquettes, increments, parallel_process_count)
            for indices, s3d in unique_3d_subtemplates.subtemplates.items()
        ]

        with Pool(processes=parallel_process_count) as pool:
            results = pool.map(_compute_detector_for_subtemplate, args_list)

        # After synchronizing all child processes, we get all computed detectors,
        # first we add them to database if it is provides, then we shift the coordinates of them
        for indices, detectors_set in results:
            subtemplates = _extract_subtemplates_from_s3d(
                unique_3d_subtemplates.subtemplates[indices]
            )
            if database is not None:
                database.add_situation(subtemplates, plaquettes, detectors_set)
            detectors_by_subtemplate[indices] = _shift_detectors_to_center_of_subtemplate(
                detectors_set, subtemplates, increments
            )

    # If parallel_process_count == 1, computing detectors sequentially
    elif parallel_process_count == 1:
        detectors_by_subtemplate = {
            indices: compute_detectors_at_end_of_situation(
                _extract_subtemplates_from_s3d(s3d),
                plaquettes,
                increments,
                database,
                only_use_database,
            )
            for indices, s3d in unique_3d_subtemplates.subtemplates.items()
        }
    # Else, invalid parallel_process_count
    else:
        raise TQECError(
            f"Invalid parallel_process_count: {parallel_process_count}. "
            "Expected a positive integer or -1 for using all available CPU cores."
        )

    # We know for sure that detectors in each subtemplate all involve a measurement
    # on at least one syndrome qubit of the central plaquette. That means that
    # detectors computed here are unique and we do not have to check for
    # duplicates.
    # Also, the last timestep template origin might not be (0, 0), so we have
    # to shift detectors accordingly.
    last_template_origin = templates[-1].instantiation_origin(k)
    detectors: list[Detector] = []
    # The below line is not strictly needed, but makes type checkers happy with
    # type inference. See https://numpy.org/doc/stable/reference/typing.html#d-arrays
    # for more information on why this should be done.
    subtemplate_indices_list: list[list[list[int]]] = [
        to2dlist(arr) for arr in unique_3d_subtemplates.subtemplate_indices
    ]
    for i, row in enumerate(subtemplate_indices_list):
        for j, subtemplate_indices in enumerate(row):
            if all(i == 0 for i in subtemplate_indices):
                continue
            detectors.extend(
                d.offset_spatially_by(
                    (j + last_template_origin.x) * increments.x,
                    (i + last_template_origin.y) * increments.y,
                )
                for d in detectors_by_subtemplate[tuple(subtemplate_indices)]
            )
    # Second filter, here to catch the duplicated detectors that were not
    # filtered by the _best_effort_filter_detectors function.
    detectors = list(set(detectors))
    return detectors
